import networkx
import pandas
import pyomo
import pyomo.opt
import pyomo.environ as pe
import scipy
import itertools
import logging

#==================================================================================
# With the following few lines of code, we are reading in the network specification.
#- We read in the set of nodes and associated node data from a nodes.csv file
#- We read in the set of arcs and their data from an arcs.csv file
#- We create a networkx object to store all of that.  This part is not strictly
# necessary -- you could go directly from the csv files to Pyomo LPs, but
# networkx is easier in terms of programming.  That ease of use and access is the
# only reason we make the networkx object.
#==================================================================================

def create_graph(nodefile, arcfile):
    g = networkx.DiGraph()
    # read the data files
    # We assume the first column of the node file is the node name
    # We also assume the arc file's first two columns are (node start, node end)
    nodes = pandas.read_csv(nodefile)
    arcs = pandas.read_csv(arcfile)
    # create the nodes and associated data
    nameCol = nodes.columns[0]
    for num,arow in nodes.iterrows():
        g.add_node( arow[nameCol], **(dict(arow))) 
    # create the arcs and associated data
    firstNodeCol = arcs.columns[0]
    secondNodeCol = arcs.columns[1]
    for num,arow in arcs.iterrows():
        g.add_edge( arow[firstNodeCol], arow[secondNodeCol], **(dict(arow))) 
    return g
    
# ============================================================================
# Lets define the typical shortest path LP.
# 
# There are a few new things about our definition of the shortest path LP.
# First, we have a parameter 'delay.'  This parameter stores the delay we have
# on an arc, if we put up a roadblock there.  This is like the 'd_ij' we talked
# about in class, except its the same for all arcs.  Based on the calculations
# from the first part of the assignment, a half-hour delay is about 110.5 units
# on the delay parameter.
# 
# Second, our definition includes the parameter xbar(i,j).  We'll use that to
# solve for Tihomir's shortest escape route, given a road block plan X.  For
# now, we'll just set the xbar(i,j) to zero.  If there is a roadblock on arc i
# j, we'll set xbar(i,j) to one.
#
# Third, we compute nCmax based on the network data.  We can use this value as
# the equivalent of infinity in terms of delay calculations.
# ============================================================================

def create_pyomo_network_lp_primal(g,edge_cost_field='cost',delay_on_field='xbar',node_rhs_field='b'):
    # Create the model
    model = pe.ConcreteModel()
    # Tell pyomo to read in dual-variable information from the solver
    model.dual = pe.Suffix(direction=pe.Suffix.IMPORT)

    # Associate the graph with this model
    model.g = g

    # Create the problem data
    model.node_set = pe.Set( initialize=g.nodes() )
    model.edge_set = pe.Set( initialize=g.edges() )
    
    model.node_rhs = pe.Param( model.node_set, 
            initialize=lambda model, n: model.g.node[n].get(node_rhs_field,0))
    model.edge_costs = pe.Param( model.edge_set, 
            initialize=lambda model, i,j: model.g.edge[i][j].get(edge_cost_field,0))
    model.edge_delay_on = pe.Param( model.edge_set, 
            initialize=lambda model, i,j: model.g.edge[i][j].get(delay_on_field,0), mutable=True)


    # Compute nCmax
    nCmax = model.g.number_of_nodes() * max([data.get(edge_cost_field, 0) for e0,e1,data in model.g.edges(data=True) ])

    model.delay = pe.Param(initialize=110.5, mutable=True)
    model.nCmax = pe.Param(initialize=nCmax)

    # Create the variables
    model.y = pe.Var(model.edge_set, domain=pe.NonNegativeReals)

    # Create the objective
    def obj_rule(model):
        return  pe.summation(model.edge_costs, model.y) + model.delay * pe.summation(model.edge_delay_on, model.y)

    model.OBJ = pe.Objective(rule=obj_rule)

    # Create the constraints, one for each node
    def flow_bal_rule(model, n):
        expr = 0
        if model.g.in_edges(n):
            expr += pe.summation(model.y, index=model.g.in_edges(n))
        if model.g.out_edges(n):
            expr += - pe.summation(model.y, index=model.g.out_edges(n))
        constr = (expr == model.node_rhs[n])
        if isinstance(constr, bool):
            if constr == True:
                return pe.Constraint.Feasible
            return pe.Constraint.Infeasible
        return constr

    model.FlowBalance = pe.Constraint(model.node_set, rule=flow_bal_rule)

    # Solve the model
    model.create()
    model= resolve_mip(model)

    # Print the model objective
    print 'Primal objective function value:', model.OBJ()
    return model

# ============================================================================
# This is a useful function, in case you want to re-solve a model in the
# interpreter.  Remember that if you dynamically want to change model
# parameters, you have to declare them as 'mutable' in the model.  If you look
# above, we've declared edge_delay_on and delay as mutable in this case.
# ============================================================================

def resolve_mip(model, tee=True):
    model.preprocess()
    solver = pyomo.opt.SolverFactory('cplex')
    results = solver.solve(model, tee=tee, keepfiles=False, options="mip_tolerances_integrality=1e-9 mip_tolerances_mipgap=0")

    # Check that we actually computed an optimal solution, load results
    if (results.solver.status != pyomo.opt.SolverStatus.ok):
        logging.warning('Check solver not ok?')
    if (results.solver.termination_condition != pyomo.opt.TerminationCondition.optimal):  
        logging.warning('Check solver optimality?')

    model.load(results)
    
    model.g.OBJ_val = model.OBJ()
    
    # Load solution data back into the networkx object 
    if hasattr(model, 'y'):
        for e in model.edge_set:
            model.g.edge[e[0]][e[1]]['flow_val'] = model.y[e].value

    if hasattr(model, 'dual'):
        for n in model.FlowBalance.iterkeys(): 
            model.g.node[n]['dual_val'] = model.dual[ model.FlowBalance[n] ]
    
    if hasattr(model, 'x'):
        for e in model.edge_set:
            model.g.edge[e[0]][e[1]]['roadblockActive'] = model.x[e].value
    
    return model


# ============================================================================
# Now, lets define the MIP for placing roadblocks.
# 
# The MIP includes a scalar, 'roadblocks', that specifies the number of
# roadblocks we are allowed to place.
# 
# We want to maximize the drug smuggler's shortest path.  So, the problem we
# want to solve is a 'max min' problem.  Like we did in class, to be able to
# put this into the solver, we'll just take the dual of the inner problem, so
# the problem becomes a 'max max'.  That is why in the following code, you'll
# see the dual of the shortest path problem.
# 
# The binary variables X(i,j), will specify whether we place a roadblock on arc
# i j.
#
# We could have defined the primal/dual pair in one function, using the same
# parameters but creating two pyomo models.  But, for clarity in terms of
# reading the code (at the expense of repeating some lines of code), we just
# create the dual on its own.
# ============================================================================

def create_pyomo_networkInterdiction_dualMIP(g,roadblocks=0,edge_cost_field='cost',node_rhs_field='b'):
    # Create the model
    model = pe.ConcreteModel()

    # Associate the graph with this model
    model.g = g

    # Create the problem data
    model.node_set = pe.Set( initialize=g.nodes() )
    model.edge_set = pe.Set( within=model.node_set*model.node_set, initialize=g.edges() )
     
    model.node_rhs = pe.Param( model.node_set, 
            initialize=lambda model, n: model.g.node[n].get(node_rhs_field,0))
    model.edge_costs = pe.Param( model.edge_set, 
            initialize=lambda model, i,j: model.g.edge[i][j].get(edge_cost_field,0))

    model.roadblocks = pe.Param(initialize=roadblocks, mutable=True)

    # Compute nCmax, also find the start and end node
    nCmax = model.g.number_of_nodes() *  max([data.get(edge_cost_field, 0) for e0,e1,data in model.g.edges(data=True) ])
    model.nCmax = pe.Param(initialize=nCmax)
    model.delay = pe.Param(initialize=110.5, mutable=True)

    model.start_node = [n for n,data in model.g.nodes(data=True) if data.get(node_rhs_field,0) == -1][0]
    model.end_node = [n for n,data in model.g.nodes(data=True) if data.get(node_rhs_field,0) == 1][0]

    # Create the variables
    model.w = pe.Var(model.node_set, domain=pe.NonNegativeReals)
    
    model.x = pe.Var(model.edge_set, domain=pe.Binary)

    # Create the objective
    def obj_rule(model):
        return  model.w[model.end_node] - model.w[model.start_node] # like w['escape'] - w['flagstaff']

    model.OBJ = pe.Objective(rule=obj_rule, sense=pe.maximize)

    # Set the x's for non-blockable arcs
    def non_blockable_rule(model, i, j):
        if model.g.edge[i][j]['blockable'] == 0:
            return model.x[(i,j)] == 0
        else:
            return pe.Constraint.Skip

    model.NonBlockableConstraint = pe.Constraint(model.edge_set, rule=non_blockable_rule)


    # Create the dual constraints, one for each edge
    def dual_const_rule(model, i,j):
        lhs = model.w[j] - model.w[i] 
        rhs = model.edge_costs[(i,j)]
        if (model.g.edge[i][j]['blockable'] == 1):
                rhs = rhs + model.delay * model.x[(i,j)]
        return lhs <= rhs

    model.DualConstraint = pe.Constraint(model.edge_set, rule=dual_const_rule)
    
    # Create the interdiction budget constraint 
    def block_limit_rule(model):
        return pe.summation(model.x) <= model.roadblocks

    model.BlockLimit = pe.Constraint(rule=block_limit_rule)

    # Solve the model
    model.create()
    model= resolve_mip(model)

    # Print the model objective
    print 'Dual objective function value:', model.OBJ()
    return model

# ============================================================================
# Here, we are printing a nice report on the solution to the screen.  You have
# to pass a networkx object, with the flow_val field loaded.
# ============================================================================

def print_solution(g):
    cur_node = "Flagstaff"

    edges = sorted(g.edges(data=True))
    for e0,e1,data in edges:
        if data.has_key('roadblockActive') and data['roadblockActive'] == 1:
            print 'Roadblock on arc %s -> %s'%(str(e0), str(e1))
    print

    while cur_node != "Escape":
        for next_node in g.successors(cur_node):
            if g.edge[cur_node][next_node]['flow_val'] == 1:
                print 'Run from %s to %s'%(cur_node, next_node)
                cur_node = next_node
                break

    print '----------'
    print 'Total cost = ', g.OBJ_val 


# ===============================================================
# Now that we've defined all our functions, lets do something....
#
# First, lets read the data. 
# ===============================================================

g = create_graph('arizona_nodes_data.csv', 'arizona_arcs_data.csv')

# ===============================================================
# Lets solve and print the drug dealer's shortest path when there are no
# roadblocks.
# ===============================================================

mPrimal = create_pyomo_network_lp_primal(g, edge_cost_field='Distance', delay_on_field='xbar', node_rhs_field='b')

print 'Drug lord path without roadblocks:\n'
print_solution(mPrimal.g)

# ===============================================================
# Lets compute the optimal roadblock place for 1 block, recompute the
# drugdealer's response, and print the solution. 
# ===============================================================

print '---------------------------------'
print '---------------------------------'
# Creates and solves the dual
mDual = create_pyomo_networkInterdiction_dualMIP(g, roadblocks=1, edge_cost_field='Distance', node_rhs_field='b')
# We have to copy the X variables from the dual into xbar, and resolve the primal
# to compute the shortest path response
for e in mDual.edge_set:
    mDual.g.edge[e[0]][e[1]]['xbar'] = mDual.x[e].value
# Now re-solve the primal to get the path
mPrimal = create_pyomo_network_lp_primal( mDual.g , edge_cost_field='Distance')

print 'Drug lord path with %d roadblocks:\n'%(mDual.roadblocks.value)
print_solution(mDual.g)

